"""
Utilities specific to Python general functionality.
"""
import logging
from dataclasses import fields, is_dataclass


def filter_kwargs(cls, kwargs, log_warnings=True):
    """
    Filters the keyword arguments to only include those that are fields of the given dataclass.
    Args:
        cls: The dataclass type to filter against.
        kwargs: A dictionary of keyword arguments.
        log_warnings: If True, logs warnings for fields not in the dataclass.
    Returns:
        A dictionary containing only the fields that are defined in the dataclass.
    """
    if not is_dataclass(cls):
        raise TypeError(f"{cls.__name__} is not a dataclass or pydantic dataclass")

    cls_fields = {f.name for f in fields(cls)}
    filtered = {}
    for k, v in kwargs.items():
        if k in cls_fields:
            filtered[k] = v
        else:
            if log_warnings:
                logging.warning(
                    f"Warning: field '{k}' not in {cls.__name__}, dropping."
                )
    return filtered



import logging
from importlib.metadata import version, PackageNotFoundError


def gito_version() -> str:
    """
    Retrieve the current version of gito.bot package.
    Returns:
        str: The version string of the gito.bot package, or "{Dev}" if not found.
    """
    try:
        return version("gito.bot")
    except PackageNotFoundError:
        logging.warning("Could not retrieve gito.bot version.")
        return "{Dev}"


class Env:
    logging_level: int = 1
    verbosity: int = 1
    gito_version: str = gito_version()
    working_folder = "."



import abc
from typing import ClassVar, Optional

import git

from ..platform_types import PlatformType
from ..shared import get_repo_base_web_url


class BaseGitPlatform(abc.ABC):
    """
    Base class for Git hosting platform interactions.
    """
    type: ClassVar[PlatformType]
    repo_base_url: str

    def __init__(self, repo: git.Repo = None, repo_base_url: str = None):
        if not repo and not repo_base_url:
            raise ValueError("Must specify repo or repo_base_url")
        if repo:
            detected_repo_base_url = get_repo_base_web_url(repo)
            if repo_base_url and detected_repo_base_url != repo_base_url:
                raise ValueError(
                    f"Provided repo_base_url '{repo_base_url}' "
                    f"does not match detected '{detected_repo_base_url}'"
                )
            repo_base_url = detected_repo_base_url
        self.repo_base_url = repo_base_url

    @abc.abstractmethod
    def is_running_in_ci(self) -> bool:
        """
        Check if the current environment is a Continuous Integration (CI) environment
        of target git platform.
        Returns:
            bool: True if running inside a CI, False otherwise.
        """

    @abc.abstractmethod
    def ci_src_branch(self) -> Optional[str]:
        """
        Get the source branch name from the CI environment.
        Returns:
            Optional[str]: The source branch name if available, None otherwise.
        """

    @abc.abstractmethod
    def create_pr_url(self, branch: str) -> Optional[str]:
        """
        Return a URL to create a pull/merge request for the given branch.
        Args:
            branch (str): The branch name for which to create the PR/MR link.
        Returns:
            Optional[str]: The URL to create the PR/MR, or None if not applicable.
        """

    @abc.abstractmethod
    def secrets_management_url(self):
        """
        Return a URL to the secrets management page of the git platform.
        Returns:
            Optional[str]: The URL to the secrets management page, or None if not applicable.
        """

    @abc.abstractmethod
    def file_url(
        self,
        file: str,
        branch="main",
        start_line: Optional[int] = None,
        end_line: Optional[int] = None,
    ) -> Optional[str]:
        """
        Return a URL to view a file in the repository.
        Args:
            file (str): The file path in the repository.
            branch (str): The branch name. Defaults to "main".
            start_line (Optional[int]): The starting line number for highlighting. Defaults to None.
            end_line (Optional[int]): The ending line number for highlighting. Defaults to None.
        Returns:
            Optional[str]: The URL to view the file, or None if not applicable.
        """



"""
Module to identify the Git platform/provider
(GitHub, GitLab, Bitbucket, etc.) for a given repository.
"""
import os
from enum import StrEnum
from typing import Optional
from functools import lru_cache
from pathlib import Path
from urllib.parse import urlparse

from git import Repo


class PlatformType(StrEnum):
    """Enumeration of supported Git provider/platform types."""
    GITHUB = "GitHub"
    GITLAB = "GitLab"
    BITBUCKET = "Bitbucket"
    GITEA = "Gitea"
    AZURE_DEVOPS = "Azure DevOps"


def identify_git_platform_by_ci_env() -> Optional[PlatformType]:
    """
    Identify the Git provider/platform type based on CI environment variables.
    """
    git_provider_ci_identifiers = {
        PlatformType.GITHUB: ["GITHUB_ACTIONS"],
        PlatformType.GITLAB: ["GITLAB_CI"],
        PlatformType.BITBUCKET: ["BITBUCKET_BUILD_NUMBER"],
    }
    for provider, env_vars in git_provider_ci_identifiers.items():
        if any(os.getenv(var) for var in env_vars):
            return provider
    return None


def identify_git_platform_from_remotes(repo_or_urls: Repo | list[str]) -> Optional[PlatformType]:
    """
    Identify the Git provider/platform type based on git remote URLs.
    """
    known_urls = {
        PlatformType.GITHUB: ["github.com"],
        PlatformType.GITLAB: ["gitlab"],
        PlatformType.BITBUCKET: ["bitbucket"],
        PlatformType.GITEA: ["gitea", "forgejo", "codeberg.org"],  # not tested yet
        PlatformType.AZURE_DEVOPS: ["dev.azure.com", "visualstudio.com"],  # not tested yet
    }
    if isinstance(repo_or_urls, Repo):
        try:
            remote_urls: list[str] = repo_or_urls.remotes.origin.urls or []
        except AttributeError:
            remote_urls = []
        remotes = [extract_base_url(i) for i in remote_urls]
    else:
        remotes = repo_or_urls
    for provider, url_parts in known_urls.items():
        if any(any(part in url for part in url_parts) for url in remotes):
            return provider
    return None


def identify_git_platform(repo: Repo) -> Optional[PlatformType]:
    """
    Identify the Git provider/platform type using multiple strategies.
    """
    return identify_git_platform_by_ci_env() or \
        identify_git_platform_from_remotes(repo) or \
        identify_git_platform_from_files(repo)


@lru_cache()
def extract_base_url(git_url: str) -> str:
    """Extract base URL from git remote URL"""
    # Handle SSH URLs (git@domain:user/repo.git)
    if git_url.startswith("git@"):
        domain = git_url.split("@")[1].split(":")[0].lower()
        return f"https://{domain}"

    parsed = urlparse(git_url)

    # Handle ssh://git@domain/... URLs
    # @todo: verify this is actual case and works as expected (not tested)
    if parsed.scheme == "ssh" and parsed.hostname:
        return f"https://{parsed.hostname}".lower()

    # Handle HTTPS URLs
    return f"{parsed.scheme}://{parsed.netloc}".lower()


def identify_git_platform_from_files(repo: Repo) -> Optional[PlatformType]:
    """
    Identify the Git provider/platform type based on provider-specific files in the repository.
    """
    if repo.working_tree_dir is None:
        return None
    git_platform_specific_files = {
        PlatformType.GITHUB: [".github"],
        PlatformType.GITLAB: [".gitlab", ".gitlab-ci.yml"],
        PlatformType.GITEA: [".gitea"],  # not tested yet
        PlatformType.AZURE_DEVOPS: ["azure-pipelines.yml", ".azure-pipelines"],  # not tested yet
    }
    path = Path(repo.working_tree_dir)
    for provider, files in git_platform_specific_files.items():
        if any((path / file).exists() for file in files):
            return provider
    return None



"""
Utilities for interacting with various Git hosting platforms.
"""
import git

from .platform_types import PlatformType, identify_git_platform
from .adapters import get_platform_adapter, BaseGitPlatform


def platform(repo: git.Repo) -> BaseGitPlatform:
    """
    Get the appropriate Git platform adapter for the given repository.
    Args:
        repo (git.Repo): The Git repository instance.
    Returns:
        BaseGitPlatform: An instance of the platform adapter.
    Raises:
        ValueError: If the platform type is unsupported or if the input type is invalid.
    """
    platform_type: PlatformType = identify_git_platform(repo)
    return get_platform_adapter(platform_type, repo=repo)



"""Allow running the package with `python -m gito`."""
# Use an absolute import (package-qualified) here; otherwise, the Windows build
# produced by PyInstaller fails.
from gito.entrypoint import main

if __name__ == "__main__":
    main()



import os
import sys
import logging

import requests
import typer

from ..cli_base import app
from ..issue_trackers import resolve_issue_key
from ..utils.git import get_cwd_repo_or_fail


class LinearAPIError(Exception):
    """Custom exception for Linear API errors."""


def post_linear_comment(issue_key: str, text: str, api_key: str) -> dict:
    """
    Post a comment to a Linear issue using the Linear API.
    Args:
        issue_key (str): The ID of the Linear issue to comment on.
        text (str): The comment text to post.
        api_key (str): The Linear API key for authentication.
    Returns:
        dict: The JSON response from the Linear API.
    Raises:
        LinearAPIError: If the API request fails for any reason.
    """
    try:
        response = requests.post(
            'https://api.linear.app/graphql',
            headers={'Authorization': api_key, 'Content-Type': 'application/json'},
            json={
                'query': '''
                    mutation($issueId: String!, $body: String!) {
                        commentCreate(input: {issueId: $issueId, body: $body}) {
                            comment { id }
                        }
                    }
                ''',
                'variables': {'issueId': issue_key, 'body': text}
            }
        )
        response.raise_for_status()
        data = response.json()

        # Check for GraphQL-level errors
        if 'errors' in data:
            raise LinearAPIError(f"GraphQL error: {data['errors']}")

        return data

    except requests.exceptions.RequestException as e:
        raise LinearAPIError(f"Request failed: {e}") from e


def _process_text_input(text: str | None) -> str:
    if text == "-":
        # Explicit stdin request
        text = sys.stdin.read()
    elif text is None:
        if not sys.stdin.isatty():
            # Data is being piped in
            text = sys.stdin.read()
    if not text or not text.strip():
        raise typer.BadParameter(
            "Comment text is required. Provide text as argument or pipe from stdin."
        )

    text = text.replace('\\n', '\n').replace('\\t', '\t')
    return text


@app.command(help="Post a comment with specified text to the associated Linear issue.")
def linear_comment(
    text: str = typer.Argument(
        default=None,
        callback=_process_text_input,
        help="Comment text (supports Markdown). Use '-' to read from stdin.",
    ),
    issue_key: str = typer.Option(
        None,
        "--issue-key",
        "-k",
        help="Linear issue key (if not provided, will be resolved from the current repo branch)",
    ),
):
    api_key = os.getenv("LINEAR_API_KEY")
    if not api_key:
        logging.error("LINEAR_API_KEY environment variable is not set")
        raise typer.Exit(code=1)

    repo = get_cwd_repo_or_fail()
    key = issue_key or resolve_issue_key(repo)
    if not key:
        logging.error("Could not determine Linear issue key from the current branch or argument")
        raise typer.Exit(code=1)
    try:
        post_linear_comment(key, text, api_key)
    except LinearAPIError as e:
        logging.error("Failed to post comment to Linear issue %s: %s", key, str(e))
        raise typer.Exit(code=1)
    logging.info("Comment posted to Linear issue %s", key)



"""
Utility functions for Git operations.
"""
import git
import typer
from microcore import ui


def get_cwd_repo_or_fail() -> git.Repo:
    """
    Get Git repository from current working directory.

    Exits with code 2 (usage error) if not inside a Git repository.
    Returns:
        git.Repo: The Git repository object.
    Raises:
        typer.Exit: If the current folder is not a Git repository.
    """
    try:
        repo = git.Repo(".", search_parent_directories=False)
        return repo
    except git.InvalidGitRepositoryError:
        ui.error(
            "Current folder is not a Git repository.\n"
            "Navigate to your repository root and run again."
        )
        raise typer.Exit(2)



"""
Utilities for processing HTML.
"""
import re


def remove_html_comments(text):
    """
    Removes all HTML comments (<!-- ... -->) from the input text.
    """
    return re.sub(r'<!--.*?-->\s*', '', text, flags=re.DOTALL)



"""
Utilities for working with GitHub platform.
"""
import os
import re
from typing import Optional

import git
from urllib.parse import quote

from ...env import Env
from .shared import get_repo_web_url


def is_running_in_github_action() -> bool:
    """
    Check if the code is running inside a GitHub Actions environment.
    Returns:
        bool: True if running in GitHub Actions, False otherwise.
    """
    return os.getenv("GITHUB_ACTIONS") == "true"


def get_gh_create_pr_link(repo_or_base_url: git.Repo | str, branch: str) -> Optional[str]:
    """
    Return a GitHub URL to create a pull request for the given branch.
    """
    branch = quote(branch, safe='')
    return get_repo_web_url(repo_or_base_url, f"/compare/{branch}?expand=1")


def get_gh_secrets_link(repo_or_base_url: git.Repo | str) -> Optional[str]:
    """
    Return a GitHub URL to manage secrets.
    """
    return get_repo_web_url(repo_or_base_url, "/settings/secrets/actions")


def gh_ci_src_branch() -> Optional[str]:
    """
    Try to get the current branch name from GitHub Actions environment variables.
    Returns:
        Optional[str]: The branch name if available, None otherwise.
    """
    # https://docs.github.com/en/actions/reference/workflows-and-actions/variables

    if branch_name := os.getenv('GITHUB_HEAD_REF'):
        # GITHUB_HEAD_REF — The source branch of a pull request
        # Push: empty
        # PR: feature-branch (just the branch name, no refs/heads/ prefix)
        return branch_name

    ref = os.getenv("GITHUB_REF", "")
    if ref.startswith("refs/heads/"):
        return ref[len("refs/heads/"):]
    return None


def get_gh_file_link(
    repo_or_base_url: git.Repo | str,
    file: str,
    branch="main",
    start_line: Optional[int] = None,
    end_line: Optional[int] = None,
) -> Optional[str]:
    branch, file = quote(branch, safe=''), quote(file, safe='/')
    url_path = f"/blob/{branch}/{file}"
    if start_line:
        url_path += f"#L{start_line}"
        if end_line and end_line != start_line:
            url_path += f"-L{end_line}"
    return get_repo_web_url(repo_or_base_url, url_path)


def detect_github_env() -> dict:
    """
    Try to detect GitHub repository/PR info from environment variables (for GitHub Actions).
    Returns a dict with github_repo, github_pr_sha, github_pr_number, github_ref, etc.
    """
    repo = os.environ.get("GITHUB_REPOSITORY", "")
    pr_sha = os.environ.get("GITHUB_SHA", "")
    pr_number = os.environ.get("GITHUB_REF", "")
    branch = ""
    ref = os.environ.get("GITHUB_REF", "")
    # Try to resolve PR head SHA if available.
    # On PRs, GITHUB_HEAD_REF/BASE_REF contain branch names.
    if "GITHUB_HEAD_REF" in os.environ:
        branch = os.environ["GITHUB_HEAD_REF"]
    elif ref.startswith("refs/heads/"):
        branch = ref[len("refs/heads/"):]
    elif ref.startswith("refs/pull/"):
        # for pull_request events
        branch = ref

    d = {
        "github_repo": repo,
        "github_pr_sha": pr_sha,
        "github_pr_number": pr_number,
        "github_branch": branch,
        "github_ref": ref,
    }
    # Fallback for local usage: try to get from git
    if not repo or repo == "octocat/Hello-World":
        git_repo = None
        try:
            git_repo = git.Repo(Env.working_folder, search_parent_directories=True)
            origin = git_repo.remotes.origin.url
            # e.g. git@github.com:Nayjest/ai-code-review.git -> Nayjest/ai-code-review
            match = re.search(r"[:/]([\w\-]+)/([\w\-\.]+?)(\.git)?$", origin)
            if match:
                d["github_repo"] = f"{match.group(1)}/{match.group(2)}"
            d["github_pr_sha"] = git_repo.head.commit.hexsha
            d["github_branch"] = (
                git_repo.active_branch.name if hasattr(git_repo, "active_branch") else ""
            )
        except Exception:
            pass
        finally:
            if git_repo:
                try:
                    git_repo.close()
                except Exception:
                    pass
    # If branch is not a commit SHA, prefer branch for links
    if d["github_branch"]:
        d["github_pr_sha_or_branch"] = d["github_branch"]
    elif d["github_pr_sha"]:
        d["github_pr_sha_or_branch"] = d["github_pr_sha"]
    else:
        d["github_pr_sha_or_branch"] = "main"
    return d



import sys
import textwrap
from typing import Optional, Callable

import typer
import microcore as mc

from .git_platform.github import is_running_in_github_action
from .git_platform.gitlab import is_running_in_gitlab_ci
from .package_metadata import version


def is_running_in_ci() -> bool:
    """
    Check if the current environment is a Continuous Integration (CI) environment.
    Returns:
        True if running in GitHub Actions or GitLab CI, False otherwise.
    """
    return is_running_in_github_action() or is_running_in_gitlab_ci()


def make_streaming_function(handler: Optional[Callable] = None) -> Callable:
    """
    Create a streaming function that processes text chunks using an optional handler.
    Used as callback for streaming LLM responses.
    Args:
        handler (Callable, optional): A function to process each text chunk before printing.
            If None, the text chunk is printed as is.
    Returns:
        Callable: A function that takes a text chunk and processes it.
    """

    def stream(text):
        if handler:
            text = handler(text)
        print(text, end='', flush=True)

    return stream


def no_subcommand(app: typer.Typer) -> bool:
    """
    Check if no subcommand was provided to the target Typer application.
    """
    return not (
        (first_arg := next((a for a in sys.argv[1:] if not a.startswith('-')), ""))
        and first_arg in (
            cmd.name or (cmd.callback.__name__.replace('_', '-') if cmd.callback else "")
            for cmd in app.registered_commands
        )
        or '--help' in sys.argv
    )


def logo(indent=2) -> str:
    """Generate Gito ASCII art logo."""
    r = mc.ui.reset

    # Character classifications
    CHAR_TYPES = {
        'shadow': set('╔═╗║╚╝╠╣╦╩╬'),
        'letter': set('█▓▒░■'),
        'version': set('⟦⟧v0123456789.'),
        'accent': set('⌬⧉▣⟁⟨⟩∘∙→⇡{⊕}'),
    }

    # Yellow → orange → red-orange → teal → cyan → blue
    COLORS = [
        (255, 220, 0),
        (255, 180, 30),
        (255, 120, 50),
        (80, 200, 180),
        (0, 230, 255),
        (30, 160, 255),
    ]

    def lerp(a, b, t):
        """Linear interpolation."""
        return int(a + (b - a) * t)

    def get_gradient_color(t, colors=COLORS):
        """Get interpolated color at position t (0-1)."""
        segment = t * (len(colors) - 1)
        idx = min(int(segment), len(colors) - 2)
        local_t = segment - idx
        c1, c2 = colors[idx], colors[idx + 1]
        return [lerp(c1[i], c2[i], local_t) for i in range(3)]

    def apply_gradient(text, row, total_rows, dim_decorations=True):
        """Apply gradient with character-based styling."""
        chars = list(text)
        non_space = [(i, c) for i, c in enumerate(chars) if c.strip()]
        if not non_space:
            return text

        start_pos, end_pos = non_space[0][0], non_space[-1][0]
        span = max(end_pos - start_pos, 1)
        row_t = row / max(total_rows - 1, 1)
        shadow_mult = 0.7 - row_t * 0.35

        result = []
        for i, char in enumerate(chars):
            if not char.strip():
                result.append(char)
                continue

            t = (i - start_pos) / span
            rgb = get_gradient_color(t)

            if char in CHAR_TYPES['shadow']:
                rgb = [int(c * shadow_mult) for c in rgb]
            elif (
                dim_decorations
                and char not in CHAR_TYPES['letter']
                and char not in CHAR_TYPES['version']
            ):
                if char in CHAR_TYPES['accent']:
                    rgb = [230, lerp(63, 108, row_t), lerp(45, 81, row_t)]
                else:
                    rgb = [lerp(70, 35, row_t), lerp(150, 80, row_t), lerp(90, 45, row_t)]

            result.append(f"\033[38;2;{rgb[0]};{rgb[1]};{rgb[2]}m{char}")

        return ''.join(result)

    lines = [
        "⇡⇡ ∘⌬ ╭──┬──▓████▓╗──◀▓██▓╗ ████████╗→─┬─▓████▓╗──/┬─╮ ⎔ /  +",
        "∙   ˊˊ│╭─■▓█╔═════╝─○──█▓╔╝ █╔═█▓╔═█║─■▓█╔═════██╗/◉╮╰-◉/ ⧉ ∙",
        "◀--┬=⟨⟨⟨⟨▓██║ ˊ <▓██▓╗ █▓║─-╚╝ █▓║ ╚╝<▓█╔╝ ▓▓╗  █▓⟩⟩⟩⟩─┬▣ --▶",
        " ∙┬┘ -─┼◉╚▓█║∘{⊕}∘█▓╔╝ █▓║─//╮ █▓║──○─╚▓█╗ ╚═╝ █▓╔╝◉┤╭─┼▣ ∙",
        "◉-╯∙∙⟁ ╰◉-╚═▓█████▓╔╝ ▓██▓╗  ╰─█▓║─→─╮∘╚═▓████▓╔═╝──╯└-∙",
        "╭──⊕─⊕──{⊕}─╚══════╝──╚═══╝─┴──╚═╝∘∘ ╰⟁ +╚═════╝ ∘∘⌬ "+f"⟦v{version()}⟧",
    ]

    total = len(lines)
    gradient_lines = [apply_gradient(line, i, total) for i, line in enumerate(lines)]
    tagline = apply_gradient("AI Code Reviewer", total // 2, total, dim_decorations=False)

    logo_text = '\n'.join(gradient_lines + [' ' * 22 + tagline + r])
    return textwrap.indent(logo_text, ' ' * indent)



import json
import logging
from dataclasses import field, asdict, is_dataclass
from datetime import datetime
from enum import StrEnum
from pathlib import Path
from typing import Optional
from pkgutil import resolve_name

import textwrap
import microcore as mc
from microcore.utils import file_link
from colorama import Fore, Style, Back
from pydantic.dataclasses import dataclass

from .constants import JSON_REPORT_FILE_NAME, HTML_TEXT_ICON, HTML_CR_COMMENT_MARKER, REFS_VALUE_ALL
from .project_config import ProjectConfig
from .utils.string import block_wrap_lr, max_line_len
from .utils.html import remove_html_comments
from .utils.python import filter_kwargs
from .utils.markdown import syntax_hint
from .utils.git_platform.platform_types import PlatformType
from .utils.git_platform.adapters import (
    get_platform_adapter,
    BaseGitPlatform
)


@dataclass
class ReviewTarget:
    git_platform_type: Optional[PlatformType] = field(default=None)
    repo_url: Optional[str] = field(default=None)
    pull_request_id: Optional[str] = field(default=None)
    what: Optional[str] = field(default=None)
    against: Optional[str] = field(default=None)
    commit_sha: Optional[str] = field(default=None)
    filters: str | list[str] | None = field(default=None)
    use_merge_base: bool = field(default=True)
    active_branch: Optional[str] = field(default=None)

    def is_full_codebase_review(self) -> bool:
        return self.what == REFS_VALUE_ALL

    def get_platform_adapter(self, raise_exceptions: bool = False) -> Optional[BaseGitPlatform]:
        try:
            return get_platform_adapter(self.git_platform_type, repo_or_base_url=self.repo_url)
        except ValueError as e:
            if raise_exceptions:
                raise e
            logging.warning(f"Could not get platform adapter: {e}")
            return None


@dataclass
class RawIssue:
    @dataclass
    class AffectedCode:
        start_line: int = field()
        end_line: Optional[int] = field(default=None)
        proposal: Optional[str] = field(default="")

    title: str = field()
    details: Optional[str] = field(default="")
    severity: Optional[int] = field(default=None)
    confidence: Optional[int] = field(default=None)
    tags: list[str] = field(default_factory=list)
    affected_lines: list[AffectedCode] = field(default_factory=list)


@dataclass
class Issue(RawIssue):
    @dataclass
    class AffectedCode(RawIssue.AffectedCode):
        file: str = field(default="")
        affected_code: str = field(default="")

        @property
        def syntax_hint(self) -> str:
            return syntax_hint(self.file)

    id: int | str = field(kw_only=True)
    file: str = field(default="")
    affected_lines: list[AffectedCode] = field(default_factory=list)

    @staticmethod
    def from_raw_issue(file: str, raw_issue: RawIssue | dict, issue_id: int | str) -> "Issue":
        if is_dataclass(raw_issue):
            raw_issue = asdict(raw_issue)
        params = filter_kwargs(Issue, raw_issue | {"file": file, "id": issue_id})
        for i, obj in enumerate(params.get("affected_lines") or []):
            d = obj if isinstance(obj, dict) else asdict(obj)
            params["affected_lines"][i] = Issue.AffectedCode(
                **filter_kwargs(Issue.AffectedCode, {"file": file} | d)
            )
        return Issue(**params)

    def code_link(self, review_target: Optional[ReviewTarget]) -> str:
        """Generate a link to the affected code in the git platform."""
        if not review_target:
            return ""
        if not (platform := review_target.get_platform_adapter(raise_exceptions=False)):
            return ""
        branch = review_target.active_branch or "main"
        if platform.is_running_in_ci() and (ci_branch := platform.ci_src_branch()):
            branch = ci_branch
        return platform.file_url(
            file=self.file,
            branch=branch,
            start_line=self.affected_lines[0].start_line if self.affected_lines else None,
            end_line=self.affected_lines[0].end_line if self.affected_lines else None,
        ) or ""


@dataclass
class ProcessingWarning:
    """
    Warning generated during code review of files
    """
    message: str = field()
    file: str | None = field(default=None)


@dataclass
class Report:
    class Format(StrEnum):
        MARKDOWN = "md"
        CLI = "cli"
        GITLAB_QUALITY_REPORT = "gitlab_quality_report"

    issues: dict[str, list[Issue]] = field(default_factory=dict)
    summary: str = field(default="")
    number_of_processed_files: int = field(default=0)
    total_issues: int = field(init=False)
    created_at: str = field(default_factory=lambda: datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    model: str = field(default_factory=lambda: mc.config().MODEL or "")
    pipeline_out: dict = field(default_factory=dict)
    processing_warnings: list[ProcessingWarning] = field(default_factory=list)
    target: Optional[ReviewTarget] = field(default=None)

    @property
    def plain_issues(self):
        return [
            issue
            for file, issues in self.issues.items()
            for issue in issues
        ]

    def register_issues(self, issues: dict[str, list[RawIssue | dict]]):
        for file, file_issues in issues.items():
            for issue in file_issues:
                self.register_issue(file, issue)

    def register_issue(self, file: str, issue: RawIssue | dict):
        if file not in self.issues:
            self.issues[file] = []
        total = len(self.plain_issues)
        self.issues[file].append(Issue.from_raw_issue(file, issue, issue_id=total + 1))
        self.total_issues = total + 1

    def __post_init__(self):
        self.total_issues = len(self.plain_issues)

    def save(self, file_name: str = ""):
        file_name = file_name or JSON_REPORT_FILE_NAME
        with open(file_name, "w") as f:
            json.dump(asdict(self), f, indent=4)
        logging.info(f"Report saved to {mc.utils.file_link(file_name)}")

    @staticmethod
    def load(file_name: str | Path = ""):
        with open(file_name or JSON_REPORT_FILE_NAME, "r") as f:
            data = json.load(f)
        data.pop("total_issues", None)
        return Report(**data)

    def render(
        self,
        config: ProjectConfig = None,
        report_format: Format = Format.MARKDOWN,
    ) -> str:
        config = config or ProjectConfig.load()
        vars = dict(
            report=self,
            ui=mc.ui,
            Fore=Fore,
            Style=Style,
            Back=Back,
            file_link=file_link,
            textwrap=textwrap,
            block_wrap_lr=block_wrap_lr,
            max_line_len=max_line_len,
            HTML_TEXT_ICON=HTML_TEXT_ICON,
            HTML_CR_COMMENT_MARKER=HTML_CR_COMMENT_MARKER,
            remove_html_comments=remove_html_comments,
            **config.prompt_vars
        )
        template = getattr(config, f"report_template_{report_format}")
        if str(template).startswith("tpl:"):
            template_file = str(template)[4:]
        else:
            template_file = None
        if str(template).startswith("fn:"):
            fn = str(template)[3:]
        else:
            fn = None

        if template_file:
            return mc.tpl(template_file, **vars)
        elif fn:
            fn = resolve_name(fn)
            return fn(**vars)
        else:
            return mc.prompt(template, **vars)

    def to_cli(self, report_format=Format.CLI):
        output = self.render(report_format=report_format)
        print("")
        print(output)
